local UH = UnderHood
local UHB = UnderHood_Bars
local SML = LibStub( "LibSharedMedia-3.0" )
local L = LibStub( "AceLocale-3.0" ):GetLocale( "UnderHood" )

local UnitChannelInfo = UnitChannelInfo
local UnitCastingInfo = UnitCastingInfo
local UnitExists = UnitExists

local Bar = {}
Bar.__index = Bar

UHB.FramePrototype = Bar

local DURATION = 0.5
local math_abs, math_cos, math_pi, math_min = math.abs, math.cos, math.pi, math.min

local function CosineInterpolate( y1, y2, mu )
	local mu2 = (1 - math_cos( mu * math_pi )) / 2
	return y1 * (1 - mu2) + y2 * mu2
end

function Bar:Create( module, config )
	local self = {}
	setmetatable( self, Bar )

	LibStub( "AceEvent-3.0" ):Embed( self )

	self.module = module
	self.config = config
	self.provider = self.module:GetProvider( self.config.provider ):Create( self )

	self.currentValue = 0
	self.targetValue = 0
	self.animationEndTime = nil
	self.color = { r = 0, g = 0, b = 0, a = 0, }

	self.frame = UH:CreateFrame( "Button", nil, "SecureUnitButtonTemplate" ) -- parented to UnderHoodParentFrame
	self.frame:SetWidth( self.config.width )
	self.frame:SetHeight( self.config.height )
	self.frame:SetScale( self.config.scale )
	self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
	self.frame:SetScript( "OnUpdate", function( this ) if this:IsVisible() then self:OnUpdate() end end )

	self.style = self.module:GetStyle( self.config.style ):Create( self )

	self:RegisterEvent( "PLAYER_REGEN_DISABLED" )
	self:RegisterEvent( "PLAYER_REGEN_ENABLED" )

	self:SetupOptions()

	self.provider:Connect()

	self:SetupWatch()
	self:UpdateInteractivity()
	self:SetupCasting()
	
	return self
end

function Bar:PLAYER_REGEN_DISABLED()
	if type( self.provider.OnEnterCombat ) == "function" then
	    self.provider:OnEnterCombat()
	end
	
	self.inCombat = true
end

function Bar:PLAYER_REGEN_ENABLED()
	self.inCombat = nil
	
	if self.needShow then self.frame:Show(); self.needShow = nil end
	if self.needHide then self.frame:Hide(); self.needHide = nil end
	
	if type( self.provider.OnLeaveCombat ) == "function" then
	    self.provider:OnLeaveCombat()
	end
end

function Bar:Hide()
	if not self:InConfigMode() then
	    if self.inCombat then
	        self.needHide = true
		else
			self.frame:Hide()
		end
	end
end

function Bar:Show()
	if self.inCombat then
	    self.needShow = true
	else
		self.frame:Show()
	end
end

function Bar:SetValue( value )
	if self.targetValue == value then return end
	
	if not self.config.animate or not self.frame:IsVisible() then
		self.currentValue = value
		self.targetValue = value

		self.style:SetValue( value )
	else
		self.targetValue = value

		if self.currentValue ~= self.targetValue then
			self.animationEndTime = GetTime() + DURATION

			self.animating = true
		end
	end
end

function Bar:SetColor( r, g, b, a )
	self.color.r = r
	self.color.g = g
	self.color.b = b
	self.color.a = a

	self.style:SetColor( r, g, b, a )
end

function Bar:SetupWatch()
	UnregisterUnitWatch( self.frame )

	if self.provider.manageVisibility then
		if self.provider:IsBarVisible() or self:InConfigMode() then
			self.frame:Show()
		else
			self.frame:Hide()
		end
	else
		local watchUnit = self.config.unit

		if type( self.provider.GetWatchUnit ) == "function" then
			watchUnit = self.provider:GetWatchUnit()
		end

		self.frame:SetAttribute( "unit", watchUnit )

		if self:InConfigMode() then
		    self.frame:Show()
		else
			RegisterUnitWatch( self.frame )
		end
	end
end

function Bar:UpdateInteractivity()
	if self.config.interactive then
		self.frame:SetAttribute( "*type1", "target" )
		self.frame:SetAttribute( "*type2", "menu" )
		self.frame:EnableMouse( true )
		self.frame:RegisterForClicks( "AnyUp" )
	else
		self.frame:SetAttribute( "*type1", nil )
		self.frame:SetAttribute( "*type2", nil )
		self.frame:RegisterForClicks()
		self.frame:EnableMouse( false )
	end

	if self.config.interactive then
		if self.config.unit == "player" then
			self.frame.menu = function( self ) ToggleDropDownMenu( 1, nil, PlayerFrameDropDown, "cursor", 0, 0 ) end
		elseif self.config.unit == "target" then
			self.frame.menu = function( self ) ToggleDropDownMenu( 1, nil, TargetFrameDropDown, "cursor", 0, 0 ) end
		elseif self.config.unit == "focus" then
			self.frame.menu = function( self ) ToggleDropDownMenu( 1, nil, FocusFrameDropDown, "cursor", 0, 0 ) end
		elseif self.config.unit == "pet" then
			self.frame.menu = function( self ) ToggleDropDownMenu( 1, nil, PetFrameDropDown, "cursor", 0, 0 ) end
		end
	else
	    self.frame.menu = nil
	end

	ClickCastFrames = ClickCastFrames or {}

	if self.config.interactive then
		ClickCastFrames[self.frame] = true
	else
		ClickCastFrames[self.frame] = nil
	end
end

function Bar:Destroy()
	self:UnregisterAllEvents()

	self.provider:Disconnect()
	self.style:Destroy()

	self.provider = nil

	if self.frames then
		for k, v in pairs( self.frames ) do
			UH:ReleaseFrame( v )
		end

		self.frames = nil
	end

	if self.frame then
		self.frame:SetAttribute( "unit", nil )

		self.frame:SetScript( "OnUpdate", nil )
		
		UnregisterUnitWatch( self.frame )

		UH:ReleaseFrame( self.frame )
	end

	self.frame = nil
end

function Bar:InConfigMode()
	return UH:InConfigMode()
end

function Bar:SetConfigMode( mode )
	if mode then
		UnregisterUnitWatch( self.frame )

		self.frame:Show()
	else
		self:SetupWatch()
	end

	self.style:Update()
end

function Bar:OnUpdate()
	local currentTime = GetTime()

	if self.casting then
	    if currentTime > self.castEndTime then
	        self.casting = nil
	        self.fadeOut = true
	        self.stopTime = currentTime
		end

		local showTime = math_min( currentTime, self.castEndTime )
		local p = (showTime - self.castStartTime) / (self.castEndTime - self.castStartTime)

		self.style:SetCastingValue( p )
	elseif self.channeling then
	    if currentTime > self.castEndTime then
	        self.channeling = nil
	        self.fadeOut = true
	        self.stopTime = currentTime
		end

		local remainingTime = self.castEndTime - currentTime
		local p = remainingTime / (self.castEndTime - self.castStartTime)

		self.style:SetCastingValue( p )
	elseif self.fadeOut then
	    local alpha

	    if self.stopTime then
	        alpha = self.stopTime - currentTime + 1
	    else
			alpha = 0
	    end

	    if alpha >= 1 then
	        alpha = 1
		end

		if alpha <= 0 then
		    self.fadeOut = nil
		    self.stopTime = nil
		    self.style:SetCastingValue( 0 )
		else
		    local r, g, b = unpack( self.castingColor )
		    self.style:SetCastingColor( r, g, b, alpha )
		end
	else
	    self.style:SetCastingValue( 0 )
	end

	if self.animating and self.currentValue and self.targetValue then
		value = CosineInterpolate( self.currentValue, self.targetValue, 1 - ((self.animationEndTime - currentTime) / DURATION) )
	    
		self.currentValue = value
		self.style:SetValue( value )

		if self.currentValue == self.targetValue then
		    self.animating = nil
		end
	end
end

function Bar:ChangeStyle( newStyleName )
	self.style:Destroy()

	if self.frames then
		for k, v in pairs( self.frames ) do
			UH:ReleaseFrame( v )
		end

		self.frames = nil
	end

	self.config.style = newStyleName
	self.style = self.module:GetStyle( self.config.style )

	if type( self.style.DefaultSettings ) == "function" then
		self.style:DefaultSettings( self.config )
	else
		self.config.styleSettings = nil
	end

	-- Adjusting frame size in case style has changed it

	if self.frame:GetWidth() ~= self.config.width then self.frame:SetWidth( self.config.width ) end
	if self.frame:GetHeight() ~= self.config.height then self.frame:SetHeight( self.config.height ) end

	self.style = self.style:Create( self )

	if type( self.style.SetupOptions ) == "function" then
		self.options.args.appearance.plugins.style = {}
		self.style:SetupOptions( self.options.args.appearance.plugins.style )
	else
		self.options.args.appearance.plugins.style = nil
	end

	-- Restoring bar color and value
	self.style:SetColor( self.color.r, self.color.g, self.color.b, self.color.a )
	self.style:SetValue( self.targetValue )

	UH:NotifyOptionsChanged()
end

function Bar:ChangeProviderOrUnit( providerName, unit )
	self.provider:Disconnect()

	if providerName then
		self.config.provider = providerName
		self.provider = self.module:GetProvider( self.config.provider )

		if type( self.provider.DefaultSettings ) == "function" then
			self.provider:DefaultSettings( self.config )
		else
			self.config.providerSettings = nil
		end

		self.provider = self.provider:Create( self )

		if not( unit ) then unit = self.config.unit end

		if not self.provider:UnitSupported( unit ) then
			for _, v in ipairs( self.provider.supportedUnits ) do
				unit = v
				break
			end
		end
	end

	if unit then
		self.config.unit = unit
	end

	self.provider:Connect()

	self:SetupWatch()
	self:SetupCasting()

	if type( self.provider.SetupOptions ) == "function" then
		self.options.args.unitAndType.plugins.provider = {}
		self.provider:SetupOptions( self.options.args.unitAndType.plugins.provider )
	else
		self.options.args.unitAndType.plugins.provider = nil
	end

	UH:NotifyOptionsChanged()
end

function Bar:SetupOptions()
	self.options = {
		name = function() return self.config.name end,
		type = "group",
		childGroups = "tab",
		args = {
			unitAndType = {
				name = L["Unit and Type"],
				type = "group",
				order = 100,
				args = {
					type = {
						name = L["Type"],
						type = "select",
						order = 1,
						values = function() return self.module:GetProvidersNameTable() end,
						get = function( info ) return self.config.provider end,
						set = function( info, value ) self:ChangeProviderOrUnit( value, nil ) end,
					},
					unit = {
						name = L["Unit"],
						type = "select",
						order = 2,
						values = function()
							local v = {}

							if not( self.provider ) then return v end

							for _, u in ipairs( self.provider.supportedUnits ) do
								v[u] = UH.OptionsHelper.Unit[u]
							end

							return v
						end,
						get = function() return self.config.unit end,
						set = function( info, value ) self:ChangeProviderOrUnit( nil, value ) end,
					},
				},
				plugins = {},
			},
			appearance = {
				name = L["Appearance"],
				type = "group",
				order = 101,
				args = {
					style = {
						name = L["Style"],
						type = "select",
						order = 1,
						values = function() return self.module:GetStylesNameTable() end,
						get = function() return self.config.style end,
						set = function( info, value ) self:ChangeStyle( value ) end,
					},
					interactive = {
						name = L["Interactive"],
						desc = L["Click to target unit, Right-click for popup menu"],
						type = "toggle",
						order = 2,
						get = function() return self.config.interactive end,
						set = function( info, value ) self.config.interactive = value; self:UpdateInteractivity() end,
					},
					animate = {
						name = L["Animate"],
						type = "toggle",
						order = 3,
						get = function() return self.config.animate end,
						set = function( info, value )
							self.config.animate = value

							if not value then
								if self.animating then
								    self.animating = nil
									self.style:SetValue( self.targetValue )
								end
							end
						end,
					},
					casting = {
					    name = L["Casting bar"],
					    type = "toggle",
					    order = 4,
					    disabled = function()
					        if type( self.style.IsCastingSupported ) ~= "function" or not self.style:IsCastingSupported() then
					            return true
							end
							
							return self.config.unit == "targettarget"
						end,
					    get = function() return self.config.castingBar end,
					    set = function( info, value )
							self.config.castingBar = value

							self:SetupCasting()
						end,
					},
				},
				plugins = {},
			},
			positionAndSize = {
				name = L["Position and Size"],
				type = "group",
				order = 102,
				args = {
					x = {
						name = L["X"],
						type = "range",
						order = 1,
						min = -800,
						max = 800,
						step = 1,
						bigStep = 1,
						get = function() return self.config.x end,
						set = function( info, value )
							self.config.x = value
							self.frame:ClearAllPoints()
							self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
						end,
					},
					y = {
						name = L["Y"],
						type = "range",
						order = 2,
						min = -800,
						max = 800,
						step = 1,
						bigStep = 1,
						get = function() return self.config.y end,
						set = function( info, value )
							self.config.y = value
							self.frame:ClearAllPoints()
							self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
						end,
					},
					width = {
						name = L["Width"],
						type = "range",
						order = 3,
						min = 1,
						max = 500,
						step = 1,
						bigStep = 1,
						get = function() return self.config.width end,
						set = function( info, value )
							self.config.width = value
							self.frame:SetWidth( value )
							self.style:Update()
						end,
					},
					height = {
						name = L["Height"],
						type = "range",
						order = 4,
						min = 1,
						max = 500,
						step = 1,
						bigStep = 1,
						get = function() return self.config.height end,
						set = function( info, value )
							self.config.height = value
							self.frame:SetHeight( value )
							self.style:Update()
						end,
					},
					scale = {
						name = L["Scale"],
						type = "range",
						order = 5,
						min = 0.01,
						max = 1,
						step = 0.01,
						bigStep = 0.01,
						get = function() return self.config.scale end,
						set = function( info, value )
							self.config.scale = value
							self.frame:SetScale( value )
							self.style:Update()
						end,
					},
				},
			},
			zOrder = {
				name = L["Z-Order"],
				type = "group",
				order = 103,
				args = {
					strata = {
						name = L["Strata"],
						type = "select",
						order = 1,
						values = UH.OptionsHelper.Strata,
						get = function() return self.config.strata end,
						set = function( info, value ) self.config.strata = value; self.frame:SetFrameStrata( value ) end,
					},
					level = {
						name = L["Level"],
						type = "range",
						order = 2,
						min = 1,
						max = 10,
						step = 1,
						get = function() return self.config.level end,
						set = function( info, value ) self.config.level = value; self.frame:SetFrameLevel( value ) end,
					},
				},
			},
		},
		plugins = {},
	}

	if type( self.provider.SetupOptions ) == "function" then
		self.options.args.unitAndType.plugins.provider = {}
		self.provider:SetupOptions( self.options.args.unitAndType.plugins.provider )
	end

	if self.style and type( self.style.SetupOptions ) == "function" then
		self.options.args.appearance.plugins.style = {}
		self.style:SetupOptions( self.options.args.appearance.plugins.style )
	end
end

function Bar:SetupCasting()
    self:UnregisterEvent( "UNIT_SPELLCAST_START" )
    self:UnregisterEvent( "UNIT_SPELLCAST_STOP" )
    self:UnregisterEvent( "UNIT_SPELLCAST_FAILED" )
    self:UnregisterEvent( "UNIT_SPELLCAST_DELAYED" )
    self:UnregisterEvent( "UNIT_SPELLCAST_INTERRUPTED" )
    self:UnregisterEvent( "UNIT_SPELLCAST_CHANNEL_START" )
    self:UnregisterEvent( "UNIT_SPELLCAST_CHANNEL_UPDATE" )
    self:UnregisterEvent( "UNIT_SPELLCAST_CHANNEL_STOP" )
	self:UnregisterEvent( "UNIT_SPELLCAST_CHANNEL_INTERRUPTED" )
	self:UnregisterEvent( "UNIT_SPELLCAST_INTERRUPTED" )

	self.casting = nil
	self.channeling = nil
	self.fadeOut = nil

	if self.config.castingBar and type( self.style.IsCastingSupported ) == "function" and self.style:IsCastingSupported() then
	    self:RegisterEvent( "UNIT_SPELLCAST_START" )
	    self:RegisterEvent( "UNIT_SPELLCAST_STOP" )
	    self:RegisterEvent( "UNIT_SPELLCAST_FAILED" )
	    self:RegisterEvent( "UNIT_SPELLCAST_DELAYED" )
	    self:RegisterEvent( "UNIT_SPELLCAST_INTERRUPTED" )
	    self:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_START" )
	    self:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_UPDATE" )
	    self:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_STOP" )
		self:RegisterEvent( "UNIT_SPELLCAST_CHANNEL_INTERRUPTED", "UNIT_SPELLCAST_INTERRUPTED" )

	    if self.config.unit == "target" then
	        self:RegisterEvent( "PLAYER_TARGET_CHANGED", "CheckAndStartSpellcast" )
		elseif self.config.unit == "focus" then
		    self:RegisterEvent( "PLAYER_FOCUS_CHANGED", "CheckAndStartSpellcast" )
		end
	else
	    self.style:SetCastingValue( 0 )
	end
end

function Bar:UNIT_SPELLCAST_START( event, unit )
	if unit ~= self.config.unit then return end
	
	local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo( unit )

	if not startTime then
	    self.style:SetCastingValue( 0 )
		return
	end

	self.castStartTime = startTime / 1000
	self.castEndTime = endTime / 1000
	self.castDelay = 0
	self.casting = true
	self.channeling = nil
	self.fadeOut = nil
	self.castingColor = UH.db.profile.colors.cast.casting

	local r, g, b = unpack( self.castingColor )

	self.style:SetCastingValue( 0 )
	self.style:SetCastingColor( r, g, b, 1 )
end

function Bar:UNIT_SPELLCAST_CHANNEL_START( event, unit )
	if unit ~= self.config.unit then return end
	
	local spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo( unit )
	
	if not startTime then
	    self.style:SetCastingValue( 0 )
		return
	end

	self.castStartTime = startTime / 1000
	self.castEndTime = endTime / 1000
	self.castDelay = 0
	self.casting = nil
	self.channeling = true
	self.fadeOut = nil
	self.castingColor = UH.db.profile.colors.cast.channeling

	local r, g, b = unpack( self.castingColor )

	self.style:SetCastingValue( 0 )
	self.style:SetCastingColor( r, g, b, 1 )
end

function Bar:UNIT_SPELLCAST_STOP( event, unit )
	if unit ~= self.config.unit then return end

	if self.casting then
		self.casting = nil
		self.fadeOut = true
		self.stopTime = GetTime()

		self.style:SetCastingValue( 1 )
	end
end

function Bar:UNIT_SPELLCAST_CHANNEL_STOP( event, unit )
	if unit ~= self.config.unit then return end

	if self.channeling then
		self.channeling = nil
		self.fadeOut = true
		self.stopTime = GetTime()

		self.style:SetCastingValue( 1 )
	end
end

function Bar:UNIT_SPELLCAST_FAILED( event, unit )
	if unit ~= self.config.unit or self.channeling then return end

	self.casting = nil
	self.channeling = nil
	self.fadeOut = true

	if not( self.stopTime ) then
	    self.stopTime = GetTime()
	end

	self.castingColor = UH.db.profile.colors.cast.failed

	local r, g, b = unpack( self.castingColor )

	self.style:SetCastingValue( 1 )
	self.style:SetCastingColor( r, g, b, 1 )
end

function Bar:UNIT_SPELLCAST_DELAYED( event, unit )
	if unit ~= self.config.unit then return end

	local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo( unit )

	if startTime then
		local oldStartTime = self.castStartTime

	    self.castStartTime = startTime / 1000
	    self.castEndTime = endTime / 1000

	    self.castDelay = (self.castDelay or 0) + (startTime - (oldStartTime or startTime))
	end
end

function Bar:UNIT_SPELLCAST_INTERRUPTED( event, unit )
	if unit ~= self.config.unit then return end

	self.casting = nil
	self.channeling = nil
	self.fadeOut = true

	if not( self.stopTime ) then
	    self.stopTime = GetTime()
	end

	self.castingColor = UH.db.profile.colors.cast.interrupted

	local r, g, b = unpack( self.castingColor )

	self.style:SetCastingValue( 1 )
	self.style:SetCastingColor( r, g, b, 1 )
end

function Bar:UNIT_SPELLCAST_CHANNEL_UPDATE( event, unit )
	if unit ~= self.config.unit then return end

	local spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo( unit )

	if startTime then
		local oldStartTime = self.castStartTime

	    self.castStartTime = startTime / 1000
	    self.castEndTime = endTime / 1000

	    self.castDelay = (self.castDelay or 0) + (startTime - (oldStartTime or startTime))
	end
end

function Bar:CheckAndStartSpellcast()
    if self.castingTimer then
        self:CancelTimer( self.castingTimer )
        self.castingTimer = nil
    end

	self.casting = nil
	self.channeling = nil
	self.fadeOut = nil

    self.style:SetCastingValue( 0 )

	if not UnitExists( self.config.unit ) then
	    return
	end

	local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo( self.config.unit )

	if spell then
	    self.castStartTime = startTime / 1000
	    self.castEndTime = endTime / 1000
	    self.castDelay = 0
	    self.casting = true
	    self.channeling = nil
	    self.fadeOut = nil
	    self.castingColor = UH.db.profile.colors.cast.casting
	else
	    spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo( self.config.unit )

	    if spell then
		    self.castStartTime = startTime / 1000
		    self.castEndTime = endTime / 1000
		    self.castDelay = 0
		    self.casting = nil
		    self.channeling = true
		    self.fadeOut = nil
		    self.castingColor = UH.db.profile.colors.cast.channeling
	    end
	end

	if self.casting or self.channeling then
		local r, g, b = unpack( self.castingColor )

		self.style:SetCastingValue( 0 )
		self.style:SetCastingColor( r, g, b, 1 )
	end
end
